import atexit
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import ClassVar

import PyQt6Ads as QtAds
from PyQt6.QtCore import QObject, QSettings, pyqtSignal

sys.path.append(str(Path(__file__).resolve().parent))
from dockindock import DockInDockWidget
from dockindockmanager import DockInDockManager, MoveDockWidgetAction

GROUP_PREFIX = "Group"


def findWidget(name, managers: "list[DockInDockManager]") -> QtAds.CDockWidget:
    for mgr in managers:
        widget = mgr.findDockWidget(name)
        if widget:
            return widget


class PerspectiveInfo:
    # Partially bypass ADS perspective management, store list here
    # and then ADS will only have one perspective loaded
    # this is because all docking widgets must exist when a perspective is loaded
    # we will guarantee that!

    settings = QSettings()
    groups: ClassVar[dict[str, list[str]]] = {}


class PerspectivesManager(QObject):
    perspectivesListChanged = pyqtSignal()
    openingPerspective = pyqtSignal()
    openedPerspective = pyqtSignal()

    def __init__(self, perspectives_folder):
        super().__init__()
        self.__perspectives_folder = perspectives_folder
        self.__perspectives = {}
        atexit.register(self.cleanup)

    def cleanup(self):
        for perspective in self.__perspectives.values():
            filename = perspective.settings.fileName()
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass

    def perspectiveNames(self) -> "list[str]":
        return self.__perspectives.keys()

    def addPerspective(self, name: str, widget: DockInDockWidget) -> None:
        if self.__perspectives_folder:
            self.__perspectives[name] = perspective = PerspectiveInfo()
            perspective.settings = self.getSettingsObject(
                self.getSettingsFileName(name, True)
            )
            perspective.groups = widget.getManager().getGroupContents()

            # save perspective internally
            widget.getManager().addPerspectiveRec(name)
            # store it in QSettings object
            widget.getManager().savePerspectivesRec(perspective.settings)
            # remove internal perspectives
            widget.getManager().removePerspectives(
                widget.getManager().perspectiveNames()
            )

        self.perspectivesListChanged.emit()

    def openPerspective(self, name: str, widget: DockInDockWidget) -> None:
        assert widget.getTopLevelDockWidget() == widget

        if self.__perspectives_folder:
            if name in self.__perspectives:
                self.openingPerspective.emit()

                if widget.canCreateNewGroups():
                    cur_groups = widget.getManager().allManagers(True, True)
                    for group in self.__perspectives[name].groups.keys():
                        found = False
                        for curgroup in cur_groups:
                            if curgroup.getPerspectiveGroupName() == group:
                                found = True
                                break
                        if not found:
                            group = DockInDockManager.getGroupNameFromPersistGroupName(
                                group
                            )

                        # restore group in file but not in GUI yet
                        widget.createGroup(group, None)

                    cur_groups = widget.getManager().allManagers(False, True)
                    for curgroup in cur_groups:
                        if (
                            curgroup.getPersistGroupName()
                            not in self.__perspectives[name].groups.keys()
                        ):
                            widget.destroyGroup(curgroup.parent())

                managers = widget.getManager().allManagers(True, True)
                for group in self.__perspectives[name].groups().keys():
                    for mgr in managers:
                        if mgr.getPersistGroupName() == group:
                            for widget_name in self.__perspectives[name].groups[group]:
                                widget = findWidget(widget_name, [mgr])
                                if widget:
                                    pass  # OK, widget is already in the good manager!
                                else:
                                    widget = findWidget(widget_name, managers)
                                    if widget:
                                        # move dock widget in the same group as it used
                                        # to be when perspective was saved this
                                        # guarantee load/open perspectives will work
                                        # smartly
                                        MoveDockWidgetAction.move(widget, mgr)

                # internally load perspectives from QSettings
                widget.getManager().loadPerspectivesRec(
                    self.__perspectives[name].settings
                )
                # load perspective (update GUI)
                widget.getManager().openPerspectiveRec(name)
                # remove internal perspectives
                widget.getManager().removePerspectives(
                    widget.getManager().perspectiveNames()
                )

                self.openedPerspective().emit()
        else:
            raise AssertionError()

    def removePerspectives(self) -> None:
        self.__perspectives.clear()
        self.perspectivesListChanged.emit()

    def removePerspective(self, name: str) -> None:
        del self.__perspectives[name]
        self.perspectivesListChanged.emit()

    def getSettingsFileName(self, perspective: str, temp: bool) -> str:
        ext = perspective + ".tmp" if temp else perspective + ".ini"
        name = "perspectives.ini" if not perspective else f"perspectives_{ext}"

        return os.path.join(self.__perspectives_folder, name)

    def getSettingsObject(self, file_path: str) -> QSettings:
        return QSettings(file_path, QSettings.Format.IniFormat)

    def loadPerspectives(self) -> None:
        if self.__perspectives_folder:
            tempfile.mktemp(dir=self.__perspectives_folder)

            self.__perspectives.clear()

            main_settings = self.getSettingsObject(self.getSettingsFileName("", False))
            main_settings.fileName()

            size = main_settings.beginReadArray("Perspectives")

            for i in range(0, size):
                main_settings.setArrayIndex(i)
                perspective = main_settings.value("Name")

                if perspective:
                    to_load = self.getSettingsFileName(perspective, False)
                    loaded = self.getSettingsFileName(perspective, True)

                    try:
                        os.remove(loaded)
                    except FileNotFoundError:
                        pass
                    if not shutil.copy(to_load, loaded):
                        raise AssertionError()

                    self.__perspectives[perspective] = PerspectiveInfo()
                    self.__perspectives[perspective].settings = self.getSettingsObject(
                        loaded
                    )

                    # load group info:
                    main_settings.beginGroup(GROUP_PREFIX)
                    for key in main_settings.allKeys():
                        self.__perspectives[perspective].groups[key] = (
                            main_settings.value(key)
                        )
                    main_settings.endGroup()
                else:
                    raise AssertionError()

            main_settings.endArray()

        self.perspectivesListChanged.emit()

    def savePerspectives(self) -> None:
        if self.__perspectives_folder:
            main_settings = self.getSettingsObject(self.getSettingsFileName("", False))

            # Save list of perspective and group organization
            main_settings.beginWriteArray("Perspectives", len(self.__perspectives))
            for i, perspective in enumerate(self.__perspectives.keys()):
                main_settings.setArrayIndex(i)
                main_settings.setValue("Name", perspective)
                main_settings.beginGroup(GROUP_PREFIX)
                for group in self.__perspectives[perspective].groups.keys():
                    main_settings.setValue(
                        group, list(self.__perspectives[perspective].groups[group])
                    )
                main_settings.endGroup()
            main_settings.endArray()

            # Save perspectives themselves
            for perspective_name in self.__perspectives.keys():
                to_save = self.getSettingsFileName(perspective_name, False)
                settings = self.__perspectives[perspective_name].settings
                settings.sync()

                try:
                    os.remove(to_save)
                except FileNotFoundError:
                    pass
                if not shutil.copy(settings.fileName(), to_save):
                    raise AssertionError()



import datetime
import logging
import os
import sys

from PyQt5 import uic
from PyQt5.QtCore import (QCoreApplication, QDir, Qt, QSettings, QSignalBlocker,
                          QRect, QPoint, qDebug, qInstallMessageHandler,
                          QtDebugMsg, QtInfoMsg, QtWarningMsg,
                          QtCriticalMsg, QtFatalMsg, QSize)
from PyQt5.QtGui import (QGuiApplication, QIcon, QCloseEvent)
from PyQt5.QtWidgets import (QCalendarWidget, QFileSystemModel, QFrame, QLabel,
                             QMenu, QTreeView, QAction, QWidgetAction,
                             QComboBox, QStyle, QSizePolicy, QInputDialog, QMenu,
                             QToolButton, QWidget, QPlainTextEdit,
                             QTableWidget, QTableWidgetItem, QApplication,
                             QMessageBox)
try:
    from PyQt5.QAxContainer import QAxWidget
except ImportError:
    ACTIVEX_AVAILABLE = False
else:
    ACTIVEX_AVAILABLE = True

from PyQtAds import QtAds

import rc  # pyrcc5 demo.qrc -o rc.py
from status_dialog import CStatusDialog

UI_FILE = os.path.join(os.path.dirname(__file__), 'mainwindow.ui')
MainWindowUI, MainWindowBase = uic.loadUiType(UI_FILE)


class _State:
    label_count = 0
    calendar_count = 0
    file_system_count = 0
    editor_count = 0
    table_count = 0
    activex_count = 0
    

def features_string(dock_widget: QtAds.CDockWidget) -> str:
    '''Function returns a features string with closable (c), movable (m) and floatable (f)
       features. i.e. The following string is for a not closable but movable and floatable
       widget: c- m+ f+'''

    f = dock_widget.features()
    closable = f & QtAds.CDockWidget.DockWidgetClosable
    movable = f & QtAds.CDockWidget.DockWidgetMovable
    floatable = f & QtAds.CDockWidget.DockWidgetFloatable
    
    return "c{} m{} f{}".format("+" if closable else "-",
                                "+" if movable else "-",
                                "+" if floatable else "-")


def append_feature_string_to_window_title(dock_widget: QtAds.CDockWidget):
    '''Appends the string returned by features_string() to the window title of
       the given DockWidget'''
       
    dock_widget.setWindowTitle(dock_widget.windowTitle() + " ({})".format(features_string(dock_widget)))


def svg_icon(filename: str):
    '''Helper function to create an SVG icon'''
    # This is a workaround, because because in item views SVG icons are not
    # properly scaled and look blurry or pixelate
    icon = QIcon(filename)
    icon.addPixmap(icon.pixmap(92))
    return icon


class CMinSizeTableWidget(QTableWidget):
    """Custom QTableWidget with a minimum size hint to test CDockWidget
    setMinimumSizeHintMode() function of CDockWidget"""
    
    def minimumSizeHint(self) -> QSize:
        return QSize(300, 100)

    
class CCustomComponentsFactory(QtAds.CDockComponentsFactory):

    def createDockAreaTitleBar(self, dock_area: QtAds.CDockAreaWidget) -> QtAds.CDockAreaTitleBar:
        title_bar = QtAds.CDockAreaTitleBar(dock_area)
        custom_button = QToolButton(dock_area)
        custom_button.setToolTip("Help")
        custom_button.setIcon(svg_icon(":/adsdemo/images/help_outline.svg"))
        custom_button.setAutoRaise(True)
        index = title_bar.indexOf(title_bar.button(QtAds.TitleBarButtonTabsMenu))
        title_bar.insertWidget(index + 1, custom_button)
        return title_bar


class MainWindow(MainWindowUI, MainWindowBase):
    save_perspective_action: QAction
    perspective_list_action: QWidgetAction
    perspective_combo_box: QComboBox
    dock_manager: QtAds.CDockManager

    def __init__(self, parent=None):
        super().__init__(parent)
        self.save_perspective_action = None
        self.perspective_list_action = None
        self.perspective_combo_box = None
        self.dock_manager = None
        self.window_title_test_dock_widget = None
        self.last_docked_editor = None

        self.setupUi(self)
        self.create_actions()
        
        # uncomment the following line if the tab close button should be
        # a QToolButton instead of a QPushButton
        # QtAds.CDockManager.setConfigFlags(QtAds.CDockManager.configFlags() | QtAds.CDockManager.TabCloseButtonIsToolButton)
        
        # uncomment the following line if you want to use opaque undocking and
        # opaque splitter resizing
        #QtAds.CDockManager.setConfigFlags(QtAds.CDockManager.DefaultOpaqueConfig)
        
        # uncomment the following line if you want a fixed tab width that does
        # not change if the visibility of the close button changes
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.RetainTabSizeWhenCloseButtonHidden, True)

        # uncomment the following line if you don't want close button on DockArea's title bar
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.DockAreaHasCloseButton, False)

        # uncomment the following line if you don't want undock button on DockArea's title bar
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.DockAreaHasUndockButton, False)

        # uncomment the following line if you don't want tabs menu button on DockArea's title bar
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.DockAreaHasTabsMenuButton, False)

        # uncomment the following line if you don't want disabled buttons to appear on DockArea's title bar
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.DockAreaHideDisabledButtons, True)

        # uncomment the following line if you want to show tabs menu button on DockArea's title bar only when there are more than one tab and at least of them has elided title
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.DockAreaDynamicTabsMenuButtonVisibility, True)

        # uncomment the following line if you want floating container to always show application title instead of active dock widget's title
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.FloatingContainerHasWidgetTitle, False)

        # uncomment the following line if you want floating container to show active dock widget's icon instead of always showing application icon
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.FloatingContainerHasWidgetIcon, True)

        # uncomment the following line if you want a central widget in the main dock container (the dock manager) without a titlebar
        # If you enable this code, you can test it in the demo with the Calendar 0
        # dock widget.
        #QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.HideSingleCentralWidgetTitleBar, True)
        
        # uncomment the following line to enable focus highlighting of the dock
        # widget that has the focus
        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.FocusHighlighting, True)
        
        # uncomment if you would like to enable an equal distribution of the
        # available size of a splitter to all contained dock widgets
        # QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.EqualSplitOnInsertion, True)

        # Now create the dock manager and its content
        self.dock_manager = QtAds.CDockManager(self)
        
        # Uncomment the following line to have the old style where the dock
        # area close button closes the active tab
        # QtAds.CDockManager.setConfigFlags(QtAds.CDockManager.DockAreaHasCloseButton
                                          # | QtAds.CDockManager.DockAreaCloseButtonClosesTab)
        self.perspective_combo_box.activated[str].connect(self.dock_manager.openPerspective)
        
        self.create_content()
        # Default window geometry - center on screen
        self.resize(1280, 720)
        self.setGeometry(QStyle.alignedRect(
            Qt.LeftToRight, Qt.AlignCenter, self.frameSize(),
            QGuiApplication.primaryScreen().availableGeometry()))
        
        # self.restore_state()
        self.restore_perspectives() 

    def create_content(self):
        # Test container docking
        dock_widget = self.create_calendar_dock_widget()
        dock_widget.setFeature(QtAds.CDockWidget.DockWidgetClosable, False)
        special_dock_area = self.dock_manager.addDockWidget(QtAds.LeftDockWidgetArea, dock_widget)
        
        # For this Special Dock Area we want to avoid dropping on the center of it (i.e. we don't want this widget to be ever tabbified):
        special_dock_area.setAllowedAreas(QtAds.OuterDockAreas)
        # special_dock_area.setAllowedAreas(QtAds.LeftDockWidgetArea | QtAds.RightDockWidgetArea) # just for testing
        
        dock_widget = self.create_long_text_label_dock_widget()
        self.window_title_test_dock_widget = dock_widget
        dock_widget.setFeature(QtAds.CDockWidget.DockWidgetFocusable, False)
        self.dock_manager.addDockWidget(QtAds.LeftDockWidgetArea, dock_widget)
        file_system_widget = self.create_file_system_tree_dock_widget()
        tool_bar = file_system_widget.createDefaultToolBar()
        tool_bar.addAction(self.actionSaveState)
        tool_bar.addAction(self.actionRestoreState)
        file_system_widget.setFeature(QtAds.CDockWidget.DockWidgetFloatable, False)
        append_feature_string_to_window_title(file_system_widget)
        self.dock_manager.addDockWidget(QtAds.BottomDockWidgetArea, file_system_widget)
        
        file_system_widget = self.create_file_system_tree_dock_widget()
        file_system_widget.setFeature(QtAds.CDockWidget.DockWidgetMovable, False)
        file_system_widget.setFeature(QtAds.CDockWidget.DockWidgetFloatable, False)
        append_feature_string_to_window_title(file_system_widget)
        
        # Test custom factory - we inject a help button into the title bar
        QtAds.CDockComponentsFactory.setFactory(CCustomComponentsFactory())
        top_dock_area = self.dock_manager.addDockWidget(QtAds.TopDockWidgetArea, file_system_widget)
        # Uncomment the next line if you would like to test the
        # HideSingleWidgetTitleBar functionality
        # top_dock_area.setDockAreaFlag(QtAds.CDockAreaWidget.HideSingleWidgetTitleBar, True)
        QtAds.CDockComponentsFactory.resetDefaultFactory()

        # We create a calendar widget and clear all flags to prevent the dock area
        # from closing
        dock_widget = self.create_calendar_dock_widget()
        dock_widget.setTabToolTip("Tab ToolTip\nHodie est dies magna")
        dock_area = self.dock_manager.addDockWidget(QtAds.CenterDockWidgetArea, dock_widget, top_dock_area)
        # Now we create a action to test resizing of DockArea widget
        action = self.menuTests.addAction("Resize {}".format(dock_widget.windowTitle()))
        def action_triggered():
            splitter = QtAds.internal.findParent(QtAds.CDockSplitter, dock_area)
            if not splitter:
                return
            # We change the sizes of the splitter that contains the Calendar 1 widget
            # to resize the dock widget
            width = splitter.width()
            splitter.setSizes([width * 2/3, width * 1/3])
        action.triggered.connect(action_triggered)

        # Now we add a custom button to the dock area title bar that will create
        # new editor widgets when clicked
        custom_button = QToolButton(dock_area)
        custom_button.setToolTip("Create Editor")
        custom_button.setIcon(svg_icon(":/adsdemo/images/plus.svg"))
        custom_button.setAutoRaise(True)
        
        title_bar = dock_area.titleBar()
        index = title_bar.indexOf(title_bar.tabBar())
        title_bar.insertWidget(index + 1, custom_button)
        def on_button_clicked():
            dock_widget = self.create_editor_widget()
            dock_widget.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
            self.dock_manager.addDockWidgetTabToArea(dock_widget, dock_area)
            dock_widget.closeRequested.connect(self.on_editor_close_requested)
        custom_button.clicked.connect(on_button_clicked)

        # Test dock area docking
        right_dock_area = self.dock_manager.addDockWidget(
            QtAds.RightDockWidgetArea,
            self.create_long_text_label_dock_widget(), top_dock_area)
        self.dock_manager.addDockWidget(
            QtAds.TopDockWidgetArea,
            self.create_long_text_label_dock_widget(), right_dock_area)
    
        bottom_dock_area = self.dock_manager.addDockWidget(
            QtAds.BottomDockWidgetArea,
            self.create_long_text_label_dock_widget(), right_dock_area)

        self.dock_manager.addDockWidget(
            QtAds.CenterDockWidgetArea,
            self.create_long_text_label_dock_widget(), right_dock_area)
        self.dock_manager.addDockWidget(
            QtAds.CenterDockWidgetArea,
            self.create_long_text_label_dock_widget(), bottom_dock_area)

        
        action = self.menuTests.addAction("Set {} Floating".format(dock_widget.windowTitle()))
        action.triggered.connect(dock_widget.setFloating)
        action = self.menuTests.addAction("Set {} As Current Tab".format(dock_widget.windowTitle()))
        action.triggered.connect(dock_widget.setAsCurrentTab)
        action = self.menuTests.addAction("Raise {}".format(dock_widget.windowTitle()))
        action.triggered.connect(dock_widget.raise_)
        
        if ACTIVEX_AVAILABLE:
            flags = self.dock_manager.configFlags()
            if flags & QtAds.CDockManager.OpaqueUndocking:
                self.dock_manager.addDockWidget(QtAds.CenterDockWidgetArea,
                                                self.create_activex_widget(), right_dock_area)
                    
        for dock_widget in self.dock_manager.dockWidgetsMap().values():
            dock_widget.viewToggled.connect(self.on_view_toggled)
            dock_widget.visibilityChanged.connect(self.on_view_visibility_changed)

    def create_actions(self):
        self.toolBar.addAction(self.actionSaveState)
        self.toolBar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.actionSaveState.setIcon(svg_icon(":/adsdemo/images/save.svg"))
        self.toolBar.addAction(self.actionRestoreState)
        self.actionRestoreState.setIcon(svg_icon(":/adsdemo/images/restore.svg"))
        
        self.save_perspective_action = QAction("Create Perspective", self)
        self.save_perspective_action.setIcon(svg_icon(":/adsdemo/images/picture_in_picture.svg"))
        self.save_perspective_action.triggered.connect(self.save_perspective)
        self.perspective_list_action = QWidgetAction(self)
        self.perspective_combo_box = QComboBox(self)
        self.perspective_combo_box.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.perspective_combo_box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.perspective_list_action.setDefaultWidget(self.perspective_combo_box)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.perspective_list_action)
        self.toolBar.addAction(self.save_perspective_action)
        
        a = self.toolBar.addAction("Create Floating Editor")
        a.setProperty("Floating", True)
        a.setToolTip("Creates floating dynamic dockable editor windows that are deleted on close")
        a.setIcon(svg_icon(":/adsdemo/images/note_add.svg"))
        a.triggered.connect(self.create_editor)
        self.menuTests.addAction(a)
        
        a = self.toolBar.addAction("Create Docked Editor")
        a.setProperty("Floating", False)
        a.setToolTip("Creates a docked editor windows that are deleted on close")
        a.setIcon(svg_icon(":/adsdemo/images/docked_editor.svg"))
        a.triggered.connect(self.create_editor)
        self.menuTests.addAction(a)

        a = self.toolBar.addAction("Create Floating Table")
        a.setToolTip("Creates floating dynamic dockable table with millions of entries")
        a.setIcon(svg_icon(":/adsdemo/images/grid_on.svg"))
        a.triggered.connect(self.create_table)
        self.menuTests.addAction(a)
        
        self.menuTests.addSeparator()
        a = self.menuTests.addAction("Show Status Dialog")
        a.triggered.connect(self.show_status_dialog)
        self.menuTests.addSeparator()
        
    def closeEvent(self, event: QCloseEvent):
        self.save_state()
        self.dock_manager.deleteLater()
        super().closeEvent(event)
        
    def on_actionSaveState_triggered(self, state: bool):
        qDebug("MainWindow::on_action_save_state_triggered")
        self.save_state()
        
    def on_actionRestoreState_triggered(self, state: bool):
        qDebug("MainWindow::on_action_restore_state_triggered")
        self.restore_state()
        
    def save_perspective(self):
        perspective_name, ok = QInputDialog.getText(self, "Save perspective",
                                                    "Enter unique name:")
        
        if ok and perspective_name:
            self.dock_manager.addPerspective(perspective_name)
            _ = QSignalBlocker(self.perspective_combo_box)
            self.perspective_combo_box.clear()
            self.perspective_combo_box.addItems(self.dock_manager.perspectiveNames())
            self.perspective_combo_box.setCurrentText(perspective_name)
            
            self.save_perspectives()
            
    def on_view_toggled(self, open: bool):
        dock_widget = self.sender()
        if dock_widget is None:
            return

        qDebug("{} view_toggled({})".format(dock_widget.objectName(), open))
        
    def on_view_visibility_changed(self, visible: bool):
        dock_widget = self.sender()
        if dock_widget is None:
            return

        # qDebug("{} visibility_changed({})".format(dock_widget.objectName(), visible))
            
    def create_editor(self):
        sender = self.sender()
        floating = sender.property("Floating")
        dock_widget = self.create_editor_widget()
        dock_widget.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
        dock_widget.setFeature(QtAds.CDockWidget.DockWidgetForceCloseWithArea, True)
        dock_widget.closeRequested.connect(self.on_editor_close_requested)
        
        if floating:
            floating_widget = self.dock_manager.addDockWidgetFloating(dock_widget)
            floating_widget.move(QPoint(20, 20))
        else:
            editor_area = self.last_docked_editor.dockAreaWidget() if self.last_docked_editor is not None else None
            if editor_area is not None:
                self.dock_manager.setConfigFlag(QtAds.CDockManager.EqualSplitOnInsertion, True)
                self.dock_manager.addDockWidget(QtAds.RightDockWidgetArea, dock_widget, editor_area)
            else:
                self.dock_manager.addDockWidget(QtAds.TopDockWidgetArea, dock_widget)
            self.last_docked_editor = dock_widget
            
    def on_editor_close_requested(self):
        dock_widget = self.sender()
        result = QMessageBox.question(self, "Close Editor",
                "Editor {} contains unsaved changes? Would you like to close it?".format(dock_widget.windowTitle()))
        if result == QMessageBox.Yes:
            dock_widget.closeDockWidget()
            
    def create_table(self):
        dock_widget = self.create_table_widget()
        dock_widget.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
        floating_widget = self.dock_manager.addDockWidgetFloating(dock_widget)
        floating_widget.move(QPoint(40, 40))
        
    def show_status_dialog(self):
        dialog = CStatusDialog(self.dock_manager)
        dialog.exec_()


    def toggle_dock_widget_window_title(self):
        title = self.window_title_test_dock_widget.windowTitle()
        i = title.find(" (Test) ")
        if i == -1:
            title += " (Test) "
        else:
            title = title[i]
        self.window_title_test_dock_widget.setWindowTitle(title)

    def save_state(self):
        '''
        Saves the dock manager state and the main window geometry
        '''
        settings = QSettings("Settings.ini", QSettings.IniFormat)
        settings.setValue("mainWindow/Geometry", self.saveGeometry())
        settings.setValue("mainWindow/State", self.saveState())
        settings.setValue("mainWindow/DockingState", self.dock_manager.saveState())

    def restore_state(self):
        '''
        Restores the dock manager state
        '''
        settings = QSettings("Settings.ini", QSettings.IniFormat)
        geom = settings.value("mainWindow/Geometry")
        if geom is not None:
            self.restoreGeometry(geom)

        state = settings.value("mainWindow/State")
        if state is not None:
            self.restoreState(state)

        state = settings.value("mainWindow/DockingState")
        if state is not None:
            self.dock_manager.restore_state(state)
            
    def save_perspectives(self):
        '''
        Save the list of perspectives
        '''
        settings = QSettings("Settings.ini", QSettings.IniFormat)
        self.dock_manager.savePerspectives(settings)

    def restore_perspectives(self):
        '''
        Restore the perspective listo of the dock manager
        '''
        settings = QSettings("Settings.ini", QSettings.IniFormat)
        self.dock_manager.loadPerspectives(settings)
        self.perspective_combo_box.clear()
        self.perspective_combo_box.addItems(self.dock_manager.perspectiveNames())

    def save_perspective(self):
        perspective_name, ok = QInputDialog.getText(self, 'Save perspective', 'Enter unique name:')
        if ok and perspective_name:
            self.dock_manager.addPerspective(perspective_name)
            _ = QSignalBlocker(self.perspective_combo_box)
            self.perspective_combo_box.clear()
            self.perspective_combo_box.addItems(self.dock_manager.perspectiveNames())
            self.perspective_combo_box.setCurrentText(perspective_name)
            self.save_perspectives()

    def create_long_text_label_dock_widget(self) -> QtAds.CDockWidget:
        label = QLabel()
        label.setWordWrap(True)
        label.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        label.setText('''Label {} {} - Lorem ipsum dolor sit amet, consectetuer
    adipiscing elit.  Aenean commodo ligula eget dolor. Aenean massa. Cum
    sociis natoque penatibus et magnis dis parturient montes, nascetur
    ridiculus mus.  Donec quam felis, ultricies nec, pellentesque eu, pretium
    quis, sem.  Nulla consequat massa quis enim. Donec pede justo, fringilla
    vel, aliquet nec, vulputate eget, arcu. In enim justo, rhoncus ut,
    imperdiet a, venenatis vitae, justo. Nullam dictum felis eu pede mollis
    pretium. Integer tincidunt. Cras dapibus. Vivamus elementum semper nisi.
    Aenean vulputate eleifend tellus. Aenean leo ligula, porttitor eu,
    consequat vitae, eleifend ac, enim. Aliquam lorem ante, dapibus in, viverra
    quis, feugiat a, tellus. Phasellus viverra nulla ut metus varius laoreet.
        '''.format(_State.label_count, datetime.datetime.now().strftime("%H:%M:%S:%f")))
        dock_widget = QtAds.CDockWidget("Label {}".format(_State.label_count))
        _State.label_count += 1
        dock_widget.setWidget(label)

        self.menuView.addAction(dock_widget.toggleViewAction())
        return dock_widget

    def create_calendar_dock_widget(self) -> QtAds.CDockWidget:
        widget = QCalendarWidget()

        dock_widget = QtAds.CDockWidget("Calendar {}".format(_State.calendar_count))
        _State.calendar_count += 1
        # The following lines are for testing the setWidget() and takeWidget()
        # functionality
        dock_widget.setWidget(widget)
        dock_widget.setWidget(widget)  # what happens if we set a widget if a widget is already set
        dock_widget.takeWidget()  # we remove the widget
        dock_widget.setWidget(widget)  # and set the widget again - there should be no error
        dock_widget.setToggleViewActionMode(QtAds.CDockWidget.ActionModeShow)
        dock_widget.setIcon(svg_icon(":/adsdemo/images/date_range.svg"))
        self.menuView.addAction(dock_widget.toggleViewAction())
        return dock_widget

    def create_file_system_tree_dock_widget(self) -> QtAds.CDockWidget:
        widget = QTreeView()
        widget.setFrameShape(QFrame.NoFrame)

        m = QFileSystemModel(widget)
        m.setRootPath(QDir.currentPath())
        widget.setModel(m)

        dock_widget = QtAds.CDockWidget("Filesystem {}".format(_State.file_system_count))
        _State.file_system_count += 1
        dock_widget.setWidget(widget)
        self.menuView.addAction(dock_widget.toggleViewAction())
        return dock_widget

    def create_editor_widget(self) -> QtAds.CDockWidget:
        widget = QPlainTextEdit()
        widget.setPlaceholderText("This is an editor. If you close the editor, it will be "
                                  "deleted. Enter your text here.")
        widget.setStyleSheet("border: none")
        dock_widget = QtAds.CDockWidget("Editor {}".format(_State.editor_count))
        _State.editor_count += 1
        dock_widget.setWidget(widget)
        dock_widget.setIcon(svg_icon(":/adsdemo/images/edit.svg"))
        dock_widget.setFeature(QtAds.CDockWidget.CustomCloseHandling, True)
        self.menuView.addAction(dock_widget.toggleViewAction())

        options_menu = QMenu(dock_widget)
        options_menu.setTitle("Options")
        options_menu.setToolTip(options_menu.title())
        options_menu.setIcon(svg_icon(":/adsdemo/images/custom-menu-button.svg"))
        menu_action = options_menu.menuAction()
        # The object name of the action will be set for the QToolButton that
        # is created in the dock area title bar. You can use this name for CSS
        # styling
        menu_action.setObjectName("options_menu")
        dock_widget.setTitleBarActions([options_menu.menuAction()])
        a = options_menu.addAction("Clear Editor")
        a.triggered.connect(widget.clear)

        return dock_widget


    def create_table_widget(self) -> QtAds.CDockWidget:
        widget = CMinSizeTableWidget()
        dock_widget = QtAds.CDockWidget("Table {}".format(_State.table_count))
        _State.table_count += 1
        COLCOUNT = 5
        ROWCOUNT = 30
        widget.setColumnCount(COLCOUNT)
        widget.setRowCount(ROWCOUNT)
        for col in range(ROWCOUNT):
            widget.setHorizontalHeaderItem(col, QTableWidgetItem("Col {}".format(col + 1)))
            for row in range(ROWCOUNT):
                widget.setItem(row, col, QTableWidgetItem("T {:}-{:}".format(row + 1, col + 1)))

        dock_widget.setWidget(widget)
        dock_widget.setIcon(svg_icon(":/adsdemo/images/grid_on.svg"))
        dock_widget.setMinimumSizeHintMode(QtAds.CDockWidget.MinimumSizeHintFromContent)
        toolbar = dock_widget.createDefaultToolBar()
        action = toolbar.addAction(svg_icon(":/adsdemo/images/fullscreen.svg"), "Toggle Fullscreen")

        def on_toggle_fullscreen():
            if dock_widget.isFullScreen():
                dock_widget.showNormal()
            else:
                dock_widget.showFullScreen()

        action.triggered.connect(on_toggle_fullscreen)
        self.menuView.addAction(dock_widget.toggleViewAction())
        return dock_widget

    def create_activex_widget(self, parent: QWidget = None) -> QtAds.CDockWidget:
        widget = QAxWidget("{6bf52a52-394a-11d3-b153-00c04f79faa6}", parent)
        dock_widget = QtAds.CDockWidget("Active X {}".format(_State.activex_count))
        _State.activex_count += 1
        dock_widget.setWidget(widget)
        self.menuView.addAction(dock_widget.toggleViewAction())
        return dock_widget


def my_message_output(type, context, msg):
    if type == QtDebugMsg:
        print("Debug: {} ({}:{}, {})".format(msg, context.file, context.line, context.function))
    elif type == QtInfoMsg:
        print("Info: {} ({}:{}, {})".format(msg, context.file, context.line, context.function))
    elif type == QtWarningMsg:
        print("Warning: {} ({}:{}, {})".format(msg, context.file, context.line, context.function))
    elif type == QtCriticalMsg:
        print("Critical: {} ({}:{}, {})".format(msg, context.file, context.line, context.function))
    elif type == QtFatalMsg:
        print("Fatal: {} ({}:{}, {})".format(msg, context.file, context.line, context.function))


if __name__ == '__main__':
    QCoreApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
    QGuiApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)
    
    with open(os.path.join(os.path.dirname(__file__), "app.css"), "r") as style_sheet_file:
        app.setStyleSheet(style_sheet_file.read())
    
    qInstallMessageHandler(my_message_output)
    qDebug("Message handler test")
    
    mw = MainWindow()
    mw.show()
    app.exec_()



import atexit
import sys
from pathlib import Path

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QApplication, QLabel, QMainWindow

sys.path.append(str(Path(__file__).parent))
from dockindock import DockInDockWidget
from perspectives import PerspectivesManager


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.perspectives_manager = PerspectivesManager("persist")
        self.resize(400, 400)
        self.dock_manager = DockInDockWidget(
            self, self.perspectives_manager, can_create_new_groups=True
        )
        self.setCentralWidget(self.dock_manager)
        self.dock_manager.attachViewMenu(self.menuBar().addMenu("View"))

        previous_dock_widget = None
        for i in range(3):
            lbl = QLabel()
            lbl.setWordWrap(True)
            lbl.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
            lbl.setText("Lorem ipsum dolor sit amet, consectetuer adipiscing elit. ")

            previous_dock_widget = self.dock_manager.addTabWidget(
                lbl, f"Top label {i}", previous_dock_widget
            )

        last_top_level_dock = previous_dock_widget

        for j in range(2):
            group_manager, _ = self.dock_manager.createGroup(
                f"Group {j}", last_top_level_dock
            )

            previous_dock_widget = None

            for i in range(3):
                # Create example content label
                # this can be any application specific widget
                lbl = QLabel()
                lbl.setWordWrap(True)
                lbl.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
                lbl.setText(
                    "Lorem ipsum dolor sit amet, consectetuer adipiscing elit. "
                )

                previous_dock_widget = group_manager.addTabWidget(
                    lbl, f"ZInner {j}/{i}", previous_dock_widget
                )

            # create sub-group
            sub_group, _ = group_manager.createGroup(
                f"SubGroup {j}", previous_dock_widget
            )
            previous_dock_widget = None
            for i in range(3):
                # Create example content label
                # this can be any application specific widget
                lbl = QLabel()
                lbl.setWordWrap(True)
                lbl.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
                lbl.setText(
                    "Lorem ipsum dolor sit amet, consectetuer adipiscing elit. "
                )

                previous_dock_widget = sub_group.addTabWidget(
                    lbl, f"SubInner {j}/{i}", previous_dock_widget
                )

        self.perspectives_manager.loadPerspectives()
        atexit.register(self.cleanup)

    def cleanup(self):
        self.perspectives_manager.savePerspectives()


if __name__ == "__main__":
    app = QApplication(sys.argv)

    w = MainWindow()
    w.show()
    app.exec()



import sys
import os

from PyQt5 import uic
from PyQt5.QtCore import Qt, QSignalBlocker
from PyQt5.QtWidgets import (QApplication, QMainWindow, QLabel, QComboBox, QTableWidget,
                             QAction, QWidgetAction, QSizePolicy, QInputDialog)
from PyQt5.QtGui import QCloseEvent
import PyQtAds as QtAds

    
UI_FILE = os.path.join(os.path.dirname(__file__), 'mainwindow.ui')
MainWindowUI, MainWindowBase = uic.loadUiType(UI_FILE)


class CMainWindow(MainWindowUI, MainWindowBase):
    def __init__(self, parent=None):
        super().__init__(parent)
        
        self.setupUi(self)

        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.OpaqueSplitterResize, True)
        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.XmlCompressionEnabled, False)
        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.FocusHighlighting, True)
        self.dock_manager = QtAds.CDockManager(self)
        
        # Set central widget
        label = QLabel()
        label.setText("This is a DockArea which is always visible, even if it does not contain any DockWidgets.")
        label.setAlignment(Qt.AlignCenter)
        central_dock_widget = QtAds.CDockWidget("CentralWidget")
        central_dock_widget.setWidget(label)
        central_dock_widget.setFeature(QtAds.CDockWidget.NoTab, True)
        central_dock_area = self.dock_manager.setCentralWidget(central_dock_widget)

        # create other dock widgets
        table = QTableWidget()
        table.setColumnCount(3)
        table.setRowCount(10)
        table_dock_widget = QtAds.CDockWidget("Table 1")
        table_dock_widget.setWidget(table)
        table_dock_widget.setMinimumSizeHintMode(QtAds.CDockWidget.MinimumSizeHintFromDockWidget)
        table_dock_widget.resize(250, 150)
        table_dock_widget.setMinimumSize(200,150)
        self.dock_manager.addDockWidgetTabToArea(table_dock_widget, central_dock_area)
        table_area = self.dock_manager.addDockWidget(QtAds.DockWidgetArea.LeftDockWidgetArea, table_dock_widget)
        self.menuView.addAction(table_dock_widget.toggleViewAction())

        table = QTableWidget()
        table.setColumnCount(5)
        table.setRowCount(1020)
        table_dock_widget = QtAds.CDockWidget("Table 2")
        table_dock_widget.setWidget(table)
        table_dock_widget.setMinimumSizeHintMode(QtAds.CDockWidget.MinimumSizeHintFromDockWidget)
        table_dock_widget.resize(250, 150)
        table_dock_widget.setMinimumSize(200,150)
        self.dock_manager.addDockWidget(QtAds.DockWidgetArea.BottomDockWidgetArea, table_dock_widget, table_area)
        self.menuView.addAction(table_dock_widget.toggleViewAction())

        properties_table = QTableWidget()
        properties_table.setColumnCount(3)
        properties_table.setRowCount(10)
        properties_dock_widget = QtAds.CDockWidget("Properties")
        properties_dock_widget.setWidget(properties_table)
        properties_dock_widget.setMinimumSizeHintMode(QtAds.CDockWidget.MinimumSizeHintFromDockWidget)
        properties_dock_widget.resize(250, 150)
        properties_dock_widget.setMinimumSize(200,150)
        self.dock_manager.addDockWidget(QtAds.DockWidgetArea.RightDockWidgetArea, properties_dock_widget, central_dock_area)
        self.menuView.addAction(properties_dock_widget.toggleViewAction())

        self.createPerspectiveUi()
        
    def createPerspectiveUi(self):
        save_perspective_action = QAction("Create Perspective", self)
        save_perspective_action.triggered.connect(self.savePerspective)
        perspective_list_action = QWidgetAction(self)
        self.perspective_combo_box = QComboBox(self)
        self.perspective_combo_box.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.perspective_combo_box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.perspective_combo_box.activated[str].connect(self.dock_manager.openPerspective)
        perspective_list_action.setDefaultWidget(self.perspective_combo_box)
        self.toolBar.addSeparator()
        self.toolBar.addAction(perspective_list_action)
        self.toolBar.addAction(save_perspective_action)
        
    def savePerspective(self):
        perspective_name, ok = QInputDialog.getText(self, "Save Perspective", "Enter unique name:")
        if not perspective_name or not ok:
            return
        
        self.dock_manager.addPerspective(perspective_name)
        blocker = QSignalBlocker(self.perspective_combo_box)
        self.perspective_combo_box.clear()
        self.perspective_combo_box.addItems(self.dock_manager.perspectiveNames())
        self.perspective_combo_box.setCurrentText(perspective_name)
        
    def closeEvent(self, event: QCloseEvent):
        # Delete dock manager here to delete all floating widgets. This ensures
        # that all top level windows of the dock manager are properly closed
        self.dock_manager.deleteLater()
        super().closeEvent(event)
        
if __name__ == '__main__':
    app = QApplication(sys.argv)
    
    w = CMainWindow()
    w.show()
    app.exec_()



import os
import sys

from pyqtbuild import PyQtBindings, PyQtProject
from sipbuild import Option

class PyQtAds(PyQtProject):
	def __init__(self):
		""" Initialise the project. """

		super().__init__()

		self.bindings_factories = [ads]

class ads(PyQtBindings):
	def __init__(self, project):
		""" Initialise the bindings. """

		super().__init__(project, 'ads')
		
	def get_options(self):
		"""Our custom options that a user can pass to sip-build."""
		options = super().get_options()
		options += [
		    Option('ads_incdir',
				   help='the directory containing the ads header file',
				   metavar='DIR'),
		    Option('ads_libdir',
				   help='the directory containing the ads library',
				   metavar='DIR'),
		    Option('ads_lib',
				   help='the ads library',
				   metavar='LIB'),
		]
		return options

	def apply_user_defaults(self, tool):
		""" Set default values for user options that haven't been set yet. """
		
		resource_file = os.path.join(self.project.root_dir,'src','ads.qrc')
		print("Adding resource file to qmake project: ", resource_file)
		self.builder_settings.append('RESOURCES += '+resource_file)
		
		if self.ads_lib is not None:
		    self.libraries.append(self.ads_lib)
		    
		if self.ads_incdir is not None:
		    self.include_dirs.append(self.ads_incdir)
		    
		if self.ads_libdir is not None:
		    self.library_dirs.append(self.ads_libdir)
		
		super().apply_user_defaults(tool)



import runpy
from pathlib import Path
from unittest.mock import patch

import pytest
from PyQt6.QtWidgets import QApplication

HERE = Path(__file__)
EXAMPLES = HERE.parent.parent / "examples"
MAINS = EXAMPLES.rglob("main.py")

app = QApplication.instance() or QApplication([])


@pytest.mark.parametrize(
    "main", MAINS, ids=lambda path: str(path.parent.relative_to(EXAMPLES))
)
def test_example(main: Path) -> None:
    with patch("PyQt6.QtWidgets.QApplication"):
        runpy.run_path(str(main), run_name="__main__")
        for widget in QApplication.topLevelWidgets():
            widget.close()
            widget.deleteLater()
        QApplication.processEvents()



import os
import sys

from PyQt5 import uic
import PyQtAds as QtAds

UI_FILE = os.path.join(os.path.dirname(__file__), 'StatusDialog.ui')
StatusDialogUI, StatusDialogBase = uic.loadUiType(UI_FILE)

class CStatusDialog(StatusDialogUI, StatusDialogBase):
    
    def __init__(self, dock_manager: QtAds.CDockManager, parent=None):
        super().__init__(parent)

        self.setupUi(self)
        
        self.dock_manager = dock_manager
        self.dock_widgets = self.dock_manager.dockWidgetsMap()
        
        for key, widget in self.dock_widgets.items():
            self.dockWidgetsComboBox.addItem(key, widget)
            
    def on_dockWidgetsComboBox_currentIndexChanged(self, index: int):
        if not isinstance(index, int):
            return
        if index < 0:
            return
        
        dock_widget = self.dockWidgetsComboBox.currentData()
        self.isClosedCheckBox.setChecked(dock_widget.isClosed())
        self.isFloatingCheckBox.setChecked(dock_widget.isFloating())
        self.tabbedCheckBox.setChecked(dock_widget.isTabbed())
        self.isCurrentTabCheckBox.setChecked(dock_widget.isCurrentTab())
        self.closableCheckBox.setChecked(dock_widget.features() & QtAds.CDockWidget.DockWidgetClosable)
        self.movableCheckBox.setChecked(dock_widget.features() & QtAds.CDockWidget.DockWidgetMovable)
        self.floatableCheckBox.setChecked(dock_widget.features() & QtAds.CDockWidget.DockWidgetFloatable)
        self.deleteOnCloseCheckBox.setChecked(dock_widget.features() & QtAds.CDockWidget.DockWidgetDeleteOnClose)
        self.customCloseHandlingCheckBox.setChecked(dock_widget.features() & QtAds.CDockWidget.CustomCloseHandling)



import sys
from pathlib import Path
from typing import TYPE_CHECKING

import PyQt6Ads as QtAds
from perspectiveactions import LoadPerspectiveAction, RemovePerspectiveAction
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QInputDialog,
    QLineEdit,
    QMenu,
    QMessageBox,
    QVBoxLayout,
    QWidget,
)

sys.path.append(str(Path(__file__).resolve().parent))
from dockindockmanager import DockInDockManager, MoveDockWidgetAction

if TYPE_CHECKING:
    from perspectives import PerspectivesManager


class DockInDockWidget(QWidget):
    def __init__(
        self,
        parent,
        perspectives_manager: "PerspectivesManager",
        can_create_new_groups: bool = False,
        top_level_widget=None,
    ):
        super().__init__(parent)

        if top_level_widget is not None:
            self.__can_create_new_groups = top_level_widget.can_create_new_groups
        else:
            self.__can_create_new_groups = can_create_new_groups
        self.__top_level_dock_widget = top_level_widget if top_level_widget else self
        self.__perspectives_manager = perspectives_manager
        self.__new_perspective_default_name: str = ""

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.__mgr = DockInDockManager(self)
        layout.addWidget(self.__mgr)

    def getManager(self) -> "DockInDockManager":
        return self.__mgr

    def getTopLevelDockWidget(self) -> "DockInDockWidget":
        return self.__top_level_dock_widget

    def canCreateNewGroups(self) -> bool:
        return self.__can_create_new_groups

    def getPerspectivesManager(self) -> "PerspectivesManager":
        return self.__perspectives_manager

    def addTabWidget(
        self, widget: QWidget, name: str, after: QtAds.CDockAreaWidget, icon=QIcon()
    ) -> QtAds.CDockAreaWidget:
        for existing in (
            self.getTopLevelDockWidget().getManager().allDockWidgets(True, True)
        ):
            if existing[1].objectName() == name:
                QMessageBox.critical(
                    self, "Error", "Name '" + name + "' already in use"
                )
                return

        dock_widget = QtAds.CDockWidget(name)
        dock_widget.setWidget(widget)
        dock_widget.setIcon(icon)

        # Add the dock widget to the top dock widget area
        return self.__mgr.addDockWidget(
            QtAds.DockWidgetArea.CenterDockWidgetArea, dock_widget, after
        )

    def isTopLevel(self) -> bool:
        return not self.objectName()

    def getGroupNameError(self, group_name: str) -> str:
        if not group_name:
            return "Group must have a non-empty name"

        dock_managers = self.__mgr.allManagers(True, True)
        for mgr in dock_managers:
            if mgr.getGroupName() == group_name:
                return "Group name '" + group_name + "' already used"

        return ""

    def createGroup(
        self, group_name: str, insert_pos: QtAds.CDockAreaWidget, icon=QIcon()
    ) -> "DockInDockWidget":
        error = self.getGroupNameError(group_name)
        if error:
            QMessageBox.critical(None, "Error", error)
            return

        child = DockInDockWidget(
            self, self.__top_level_dock_widget, self.__perspectives_manager
        )
        child.setObjectName(group_name)

        dock_widget = QtAds.CDockWidget(group_name)
        dock_widget.setWidget(child)
        dock_widget.setIcon(icon)

        insert_pos = self.__mgr.addDockWidget(
            QtAds.DockWidgetArea.CenterDockWidgetArea, dock_widget, insert_pos
        )

        return child, insert_pos

    def destroyGroup(self, widget_to_remove: "DockInDockWidget") -> None:
        top_level_widget = widget_to_remove.getTopLevelDockWidget()

        if top_level_widget and top_level_widget != widget_to_remove:
            for dock_widget in (
                widget_to_remove.getManager().getWidgetsInGUIOrder()
            ):  # don't use allDockWidgets to preserve sub-groups
                MoveDockWidgetAction.move(dock_widget, top_level_widget.getManager())
            assert not widget_to_remove.getManager().allDockWidgets(True, True)

            # find widget's parent:
            for dockwidget in top_level_widget.getManager().allDockWidgets(True, True):
                if dockwidget[1].widget() == widget_to_remove:
                    dockwidget[0].removeDockWidget(dockwidget[1])
                    del dockwidget[1]
                    # delete widgetToRemove;
                    # automatically deleted when dockWidget is deleted
                    widget_to_remove = None
                    break

            assert widget_to_remove is None
        else:
            raise AssertionError()

    def attachViewMenu(self, menu: QMenu) -> None:
        menu.aboutToShow.connect(self.autoFillAttachedViewMenu)

    def autoFillAttachedViewMenu(self) -> None:
        menu = self.sender()

        if menu:
            menu.clear()
            self.setupViewMenu(menu)
        else:
            raise AssertionError()

    def setupViewMenu(self, menu):
        dock_managers = self.__mgr.allManagers(True, True)

        has_perspectives_menu = False
        if self.getTopLevelDockWidget() == self:
            has_perspectives_menu = self.__perspectives_manager is not None
        else:
            raise AssertionError()

        organize = menu
        if has_perspectives_menu:
            organize = menu.addMenu("Organize")

        self.setupMenu(organize, dock_managers)

        if has_perspectives_menu:
            perspectives = menu.addMenu("Perspectives")
            self.fillPerspectivesMenu(perspectives)

    def setupMenu(self, menu: QMenu, move_to: "list[DockInDockManager]") -> None:
        self.__mgr.fillViewMenu(menu, move_to)
        menu.addSeparator()
        move_menu = menu.addMenu("Move")
        self.__mgr.fillMoveMenu(move_menu, move_to)

    def fillPerspectivesMenu(self, menu: QMenu):
        menu.addAction("Create perspective...", self.createPerspective)
        perspectives_names = []
        if self.__perspectives_manager:
            perspectives_names = self.__perspectives_manager.perspectiveNames()

        if perspectives_names:
            load = menu.addMenu("Load perspective")
            for name in perspectives_names:
                load.addAction(LoadPerspectiveAction(load, name, self))
            remove = menu.addMenu("Remove perspective")
            for name in perspectives_names:
                remove.addAction(RemovePerspectiveAction(remove, name, self))

    def setNewPerspectiveDefaultName(self, default_name: str) -> None:
        self.__new_perspective_default_name = default_name

    def createPerspective(self) -> None:
        if not self.__perspectives_manager:
            return

        name = self.__new_perspective_default_name
        if self.__new_perspective_default_name:
            index = 2
            while name in self.__perspectives_manager.perspectiveNames():
                name = f"{self.__new_perspective_default_name}({index})"
                index += 1

        while True:
            name, ok = QInputDialog.getText(
                None,
                "Create perspective",
                "Enter perspective name",
                QLineEdit.Normal,
                name,
            )
            if ok:
                if not name:
                    QMessageBox.critical(
                        None, "Error", "Perspective name cannot be empty"
                    )
                    continue
                elif name in self.__perspectives_manager.perspectiveNames():
                    if (
                        QMessageBox.critical(
                            None,
                            "Error",
                            f"Perspective '{name}' already exists, overwrite it?",
                            QMessageBox.Yes | QMessageBox.No,
                            QMessageBox.No,
                        )
                        == QMessageBox.No
                    ):
                        continue

                self.__perspectives_manager.addPerspective(name, self)
                break
            else:
                break

    def dumpStatus(
        self,
        echo: callable = print,
        widget: QtAds.CDockWidget = None,
        tab: str = "",
        suffix: str = "",
    ) -> str:
        if widget is not None:
            as_mgr = DockInDockManager.dockInAManager(widget)
            if as_mgr:
                as_mgr.parent().dumpStatus(tab=tab)
            else:
                echo(tab + widget.objectName() + suffix)
        else:
            echo(tab + "Group: " + self.getManager().getGroupName())
            tab += "   "
            visible_widgets = set()
            for widget in self.getManager().getWidgetsInGUIOrder():
                visible_widgets.add(widget)
                self.dumpStatus(widget=widget, tab=tab)

            for closed in self.getManager().dockWidgetsMap().values():
                if closed not in visible_widgets:
                    self.dumpStatus(widget=closed, tab=tab, suffix=" (closed)")



import os
from pathlib import Path
import subprocess

from pyqtbuild import PyQtBindings, PyQtProject

ROOT = Path(__file__).parent


class PyQt6Ads(PyQtProject):
    def __init__(self):
        super().__init__()
        self.bindings_factories = [PyQt6Adsmod]

    def apply_user_defaults(self, tool):
        if tool == "sdist":
            return super().apply_user_defaults(tool)
        qmake_path = "bin/qmake"
        if os.name == "nt":
            qmake_path += ".exe"
        try:
            qmake_bin = str(next(ROOT.rglob(qmake_path)).absolute())
        except StopIteration:
            raise RuntimeError(
                "qmake not found.\n"
                "Please run `uvx --from aqtinstall aqt install-qt <plat> "
                "desktop <qtversion> <arch> --outputdir Qt`"
            )
        self.builder.qmake = qmake_bin
        return super().apply_user_defaults(tool)

    def build_wheel(self, wheel_directory):
        # use lowercase name for wheel, for
        # https://packaging.python.org/en/latest/specifications/binary-distribution-format/
        self.name = self.name.lower()
        return super().build_wheel(wheel_directory)


class PyQt6Adsmod(PyQtBindings):
    def __init__(self, project):
        super().__init__(project, "PyQt6Ads")

    def apply_user_defaults(self, tool):
        resource_file = os.path.join(
            self.project.root_dir, "Qt-Advanced-Docking-System", "src", "ads.qrc"
        )
        self.builder_settings.append("RESOURCES += " + resource_file)
        super().apply_user_defaults(tool)



import os
import sys

import PyQt6Ads as QtAds
from PyQt6 import uic
from PyQt6.QtCore import QSignalBlocker, Qt
from PyQt6.QtGui import QAction, QCloseEvent
from PyQt6.QtWidgets import (
    QApplication,
    QComboBox,
    QInputDialog,
    QLabel,
    QSizePolicy,
    QTableWidget,
    QWidgetAction,
)

UI_FILE = os.path.join(os.path.dirname(__file__), "mainwindow.ui")
MainWindowUI, MainWindowBase = uic.loadUiType(UI_FILE)


class CMainWindow(MainWindowUI, MainWindowBase):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setupUi(self)

        QtAds.CDockManager.setConfigFlag(
            QtAds.CDockManager.eConfigFlag.OpaqueSplitterResize, True
        )
        QtAds.CDockManager.setConfigFlag(
            QtAds.CDockManager.eConfigFlag.XmlCompressionEnabled, False
        )
        QtAds.CDockManager.setConfigFlag(
            QtAds.CDockManager.eConfigFlag.FocusHighlighting, True
        )
        self.dock_manager = QtAds.CDockManager(self)

        # Set central widget
        label = QLabel()
        label.setText(
            "This is a DockArea which is always visible, even if it does not "
            "contain any DockWidgets."
        )
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_dock_widget = QtAds.CDockWidget("CentralWidget")
        central_dock_widget.setWidget(label)
        central_dock_widget.setFeature(QtAds.CDockWidget.DockWidgetFeature.NoTab, True)
        central_dock_area = self.dock_manager.setCentralWidget(central_dock_widget)

        # create other dock widgets
        table = QTableWidget()
        table.setColumnCount(3)
        table.setRowCount(10)
        table_dock_widget = QtAds.CDockWidget("Table 1")
        table_dock_widget.setWidget(table)
        table_dock_widget.setMinimumSizeHintMode(
            QtAds.CDockWidget.eMinimumSizeHintMode.MinimumSizeHintFromDockWidget
        )
        table_dock_widget.resize(250, 150)
        table_dock_widget.setMinimumSize(200, 150)
        self.dock_manager.addDockWidgetTabToArea(table_dock_widget, central_dock_area)
        table_area = self.dock_manager.addDockWidget(
            QtAds.DockWidgetArea.LeftDockWidgetArea, table_dock_widget
        )
        self.menuView.addAction(table_dock_widget.toggleViewAction())

        table = QTableWidget()
        table.setColumnCount(5)
        table.setRowCount(1020)
        table_dock_widget = QtAds.CDockWidget("Table 2")
        table_dock_widget.setWidget(table)
        table_dock_widget.setMinimumSizeHintMode(
            QtAds.CDockWidget.eMinimumSizeHintMode.MinimumSizeHintFromDockWidget
        )
        table_dock_widget.resize(250, 150)
        table_dock_widget.setMinimumSize(200, 150)
        self.dock_manager.addDockWidget(
            QtAds.DockWidgetArea.BottomDockWidgetArea, table_dock_widget, table_area
        )
        self.menuView.addAction(table_dock_widget.toggleViewAction())

        properties_table = QTableWidget()
        properties_table.setColumnCount(3)
        properties_table.setRowCount(10)
        properties_dock_widget = QtAds.CDockWidget("Properties")
        properties_dock_widget.setWidget(properties_table)
        properties_dock_widget.setMinimumSizeHintMode(
            QtAds.CDockWidget.eMinimumSizeHintMode.MinimumSizeHintFromDockWidget
        )
        properties_dock_widget.resize(250, 150)
        properties_dock_widget.setMinimumSize(200, 150)
        self.dock_manager.addDockWidget(
            QtAds.DockWidgetArea.RightDockWidgetArea,
            properties_dock_widget,
            central_dock_area,
        )
        self.menuView.addAction(properties_dock_widget.toggleViewAction())

        self.createPerspectiveUi()

    def createPerspectiveUi(self):
        save_perspective_action = QAction("Create Perspective", self)
        save_perspective_action.triggered.connect(self.savePerspective)
        perspective_list_action = QWidgetAction(self)
        self.perspective_combo_box = QComboBox(self)
        self.perspective_combo_box.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToContents
        )
        self.perspective_combo_box.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        self.perspective_combo_box.currentTextChanged.connect(
            self.dock_manager.openPerspective
        )
        perspective_list_action.setDefaultWidget(self.perspective_combo_box)
        self.toolBar.addSeparator()
        self.toolBar.addAction(perspective_list_action)
        self.toolBar.addAction(save_perspective_action)

    def savePerspective(self):
        perspective_name, ok = QInputDialog.getText(
            self, "Save Perspective", "Enter unique name:"
        )
        if not perspective_name or not ok:
            return

        self.dock_manager.addPerspective(perspective_name)
        QSignalBlocker(self.perspective_combo_box)
        self.perspective_combo_box.clear()
        self.perspective_combo_box.addItems(self.dock_manager.perspectiveNames())
        self.perspective_combo_box.setCurrentText(perspective_name)

    def closeEvent(self, event: QCloseEvent):
        # Delete dock manager here to delete all floating widgets. This ensures
        # that all top level windows of the dock manager are properly closed
        self.dock_manager.deleteLater()
        super().closeEvent(event)


if __name__ == "__main__":
    app = QApplication(sys.argv)

    w = CMainWindow()
    w.show()
    app.exec()



from PyQt5.QtWidgets import QAction, QMenu


class LoadPerspectiveAction(QAction):
    def __init__(self, parent: QMenu, name: str, dock_manager: 'DockInDockWidget'):
        super().__init__(name, parent)
        self.name = name
        self.dock_manager = dock_manager
        
        self.triggered.connect(self.load)
        
    def load(self):
        self.dock_manager.getPerspectivesManager().openPerspective(self.name, self.dock_manager)
        
        
class RemovePerspectiveAction(QAction):
    def __init__(self, parent: QMenu, name: str, dock_manager: 'DockInDockWidget'):
        super().__init__(name, parent)
        self.name = name
        self.dock_manager = dock_manager
        
        self.triggered.connect(self.remove)
        
    def remove(self):
        self.dock_manager.getPerspectivesManager().removePerspective(self.name)


import sys

import PyQt6Ads as QtAds
from PyQt6.QtCore import qDebug
from PyQt6.QtGui import QAction, QCloseEvent
from PyQt6.QtWidgets import QApplication, QMainWindow, QTextEdit


class MainWindow(QMainWindow):
    dock_manager = None

    def closeEvent(self, event: QCloseEvent):
        super().closeEvent(event)
        if self.dock_manager is not None:
            self.dock_manager.deleteLater()

    def setDockManager(self, dock_manager: QtAds.CDockManager):
        self.dock_manager = dock_manager


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = MainWindow()

    QtAds.CDockManager.setConfigFlag(
        QtAds.CDockManager.eConfigFlag.FocusHighlighting, True
    )
    QtAds.CDockManager.setConfigFlag(
        QtAds.CDockManager.eConfigFlag.AllTabsHaveCloseButton, True
    )
    dock_manager = QtAds.CDockManager(w)
    w.setDockManager(dock_manager)

    count = 0

    def on_focused_dock_widget_changed(old: QtAds.CDockWidget, now: QtAds.CDockWidget):
        global count
        old_name = old.objectName() if old else "-"
        msg = (
            f"{count:d} CDockManager::focusedDockWidgetChanged old: {old_name} now: "
            f"{now.objectName()} visible: {now.isVisible()}"
        )
        qDebug(msg)
        count += 1
        now.widget().setFocus()

    dock_manager.focusedDockWidgetChanged.connect(on_focused_dock_widget_changed)

    action = QAction("New Delete On Close", w)
    w.menuBar().addAction(action)

    i = 0

    def on_action_triggered():
        global i
        dw = QtAds.CDockWidget(f"test doc {i:d}")
        i += 1
        editor = QTextEdit("lorem ipsum...", dw)
        dw.setWidget(editor)
        dw.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
        area = dock_manager.addDockWidgetTab(
            QtAds.DockWidgetArea.CenterDockWidgetArea, dw
        )
        qDebug(f"doc dock widget created! {dw} {area}")

    action.triggered.connect(on_action_triggered)

    action = QAction("New", w)
    w.menuBar().addAction(action)

    def on_action2_triggered():
        global i
        dw = QtAds.CDockWidget(f"test {i:d}")
        i += 1
        editor = QTextEdit("lorem ipsum...", dw)
        dw.setWidget(editor)
        area = dock_manager.addDockWidgetTab(
            QtAds.DockWidgetArea.CenterDockWidgetArea, dw
        )
        qDebug(f"dock widget created! {dw} {area}")

    action.triggered.connect(on_action2_triggered)

    w.show()
    app.exec()
